/** skeletons.h
 *
 *  generic functions for dataparallel skeletons 
 *  map, reduce, prefix
 *
 *  C. Kessler, reworked 991202
 */

// generic non-nestable map routine for unary functions:
sync void map( sh straight void (*f)(void *, void *), sh void **x,
               sh void **d, sh int N, sh int elsize );

// generic non-nestable map routine for binary functions with 1 scalar opnd:
sync void map1( sh straight void (*f)(void *, void *, void *),
                sh void **x, sh void **d, sh void **d2,
                sh int N, sh int elsize );

// generic non-nestable map routine for binary functions:
sync void map2( sh straight void (*f)(void *, void *, void *), 
                sh void **x, sh void **d, sh void **d2,
                sh int N, sh int elsize );

// generic nestable map routine for unary functions:
sync void Map( sync sh void (*f)(sh void *, sh void *, sh int), 
               sh void **x, sh void **d, 
               sh int N, sh int mx, sh int md, sh int elsize );

// nestable map skeleton for binary functions
// Set md2 to zero if d2 denotes a scalar value.

sync void Map2( sync sh void (*f)(sh void *, sh void *, sh void *, sh int),
                sh void **x, sh void **d, sh void **d2, 
                sh int N, sh int mx, sh int md, sh int md2, sh int elsize );

/** In-place f-reduction for an array d of n<=2p items of size elsize.
 *  Executed by p processors numbered consecutively from 0 to p-1.
 */
sync void preduce( sh straight void (*f)(void *, void *, void *),
                   sh void **d, sh int n, sh int elsize );

void seq_reduce( straight void (*f)(void *, void *, void *),
                 void *s, void **d, int n, int elsize );

sync void reduce( straight void (*f)(void *, void *, void *),
                  sh void *s, sh void **d, sh int n, sh int elsize );

/** In-place iterative parallel f-prefix computation
 *  for a p-element array with p processors.
 *  If non--null, offset is f-combined with every element.
 *  Consecutive numbering of $ in 0..groupsize()-1 is assumed.
 */
sync void pprefix( sync sh void (*f)(void *, void *, void *),
                    sh void **x, sh void *offset, sh int elsize );

sync void prefix( sync sh void (*f)(void *, void *, void *),
                  sh void **x, sh void **d,
                  sh int n, sh int elsize );

